// Copyright (C) 2012 The Libphonenumber Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "phonenumbers/string_byte_sink.h"

#include <string>

using std::string;

namespace i18n {
namespace phonenumbers {

StringByteSink::StringByteSink(string* dest) : dest_(dest) {}

StringByteSink::~StringByteSink() {}

void StringByteSink::Append(const char* data, int32_t n) {
  dest_->append(data, n);
}

}  // namespace phonenumbers
}  // namespace i18n
